
'===============================================================================
' Autor ............. Jrgen Schlottke, Schnaich-C.-Str. 46, D-W 2200 Elmshorn
' Zweck ............. Demonstration der Schachprogrammierung unter Turbo-Pascal
' Datum ....................... irgendwann 1991, als PD freigegeben am 18.01.93
' Version ................................................. ohne Versionsnummer
' FreeBasic Bearbeitung .................. Roland Chastain, Sarreguemines, 2013
'===============================================================================

' FreeBASIC 0.90.0

type brt
  as integer brett(-10 to 109)      ' Spielfeld
  as integer epfeld(-1 to 1)        ' En-Passant-Felder
  as integer zuglistevon(1 to 100)  ' Ausgangsfeld
  as integer zuglistenach(1 to 100) ' Endfeld
  as integer zugwert(1 to 100)      ' Bewertung
  as integer anz                    ' Anzahl der gefundenen Zge
  as integer material               ' Materialbilanz
  as integer farbe                  ' Farbe des am Zug befindlichen Spielers
end type

const ohnewert     =   0 ' Wert eines leeren Felds
const bauernwert   =   2 ' Wert eines Bauernfelds
const laeuferwert  =   6 ' etc.
const springerwert =   7
const turmwert     =  10
const damenwert    =  19
const koenigswert  = 126
const ausserhalb   = 127 ' Wert eines Felds auerhalb des Brettes

type tSchachprogramm
  as integer halbzugzahl = 0 ' Anzahl der gemachten Zge
  as integer rechentiefe = 3 ' Rechentiefe des Computers in Halbzgen
  as integer maxtiefe    = 5 ' Maximale Rechentiefe bei Schlagzgen
' Bewegungsrichtungen der Figuren auf dem Spielfeld
  as integer bauer   (1 to 4) = {+01,+02,-09,+11}
  as integer springer(1 to 8) = {+12,+21,+19,+08,-12,-21,-19,-08}
  as integer laeufer (1 to 4) = {+11,-11,+09,-09}
  as integer turm    (1 to 4) = {-01,+01,-10,+10}
  as integer koenig  (1 to 8) = {-01,+01,-10,+10,+11,-11,+09,-09}
  as brt spielbrett               ' Auf diesem Brett wird gespielt
  as brt grundbrett               ' Ein Brett in Grundstellung
  as integer wert, eom
  as integer varianten            ' Anzahl der berechneten Zugvarianten
  as integer mgzugvon(1 to 100)   ' Die bestmglichen Zge
  as integer mgzugnach(1 to 100)
  as integer mgzugwert(1 to 100)
  declare function Stellungswert(byref b as brt, byval farbe as integer) as integer
  declare sub Grundstellung(byref s as string = "")
  declare function Brettdarstellung as string
  declare sub Setzen(byref b as brt, byval von as integer, byval nach as integer)
  declare sub Eintragen(byref b as brt, byval von as integer, byval nach as integer)
  declare sub Zuggenerator_ohne(byref b as brt)
  declare function Schach(byref b as brt) as integer
  declare sub Rochade_testen(byref b as brt, byval vonkoenig as integer, byval schritt as integer, byval turmfeld as integer)
  declare sub Zuggenerator(byref b as brt)
  declare function Maxwertung(byref b as brt, byval farbe as integer, byval tiefe as integer, byval alfa as integer) as integer
  declare function Spielereingabe(byref eingabe as string) as integer
  declare function Zugausgeben(byval von as integer, byval nach as integer) as string
  declare function Feld_ist_in_grundstellung(byref b as brt, byval feldnr as integer) as integer
  declare function Besterzug as integer
  declare function Computerzug as string
end type

function tSchachprogramm.Stellungswert(byref b as brt, byval farbe as integer) as integer
' Zweck: Bewertung der aktuellen Brettstellung
  return b.material * farbe
end function ' Stellungswert

data -010,-007,-006,-019,-126,-006,-007,-010
data -002,-002,-002,-002,-002,-002,-002,-002
data -000,-000,-000,-000,-000,-000,-000,-000
data -000,-000,-000,-000,-000,-000,-000,-000
data -000,-000,-000,-000,-000,-000,-000,-000
data -000,-000,-000,-000,-000,-000,-000,-000
data +002,+002,+002,+002,+002,+002,+002,+002
data +010,+007,+006,+019,+126,+006,+007,+010

sub tSchachprogramm.Grundstellung(byref s as string = "")
' Zweck: Grundaufstellung der Figuren
  dim as integer x, y, z
  
  with grundbrett
    for z = -10 to 109
      .brett(z) = ausserhalb
    next z
    restore
    for y = 8 to 1 step -1
      for x = 1 to 8
        read .brett(10 * x + y)
      next x
    next y
    .epfeld(+1) = 0
    .epfeld(-1) = 0
    .material = 0
    .farbe = 1
  end with
  
  spielbrett = grundbrett
  halbzugzahl = 0
  
  if len(s) > 0 then
    
    dim as string symbol = "kqrbnp PNBRQK", s1 = s, a(1 to 6)
    
    z = instr(s1, " ") : a(1) = left(s1, z - 1) : s1 = right(s1, len(s1) - z)
    z = instr(s1, " ") : a(2) = left(s1, z - 1) : s1 = right(s1, len(s1) - z)
    z = instr(s1, " ") : a(3) = left(s1, z - 1) : s1 = right(s1, len(s1) - z)
    z = instr(s1, " ") : a(4) = left(s1, z - 1) : s1 = right(s1, len(s1) - z)
    z = instr(s1, " ") : a(5) = left(s1, z - 1) : s1 = right(s1, len(s1) - z)
    a(6) = s1
    
    with spielbrett
      x = 1
      y = 8
      for z = 1 to len(a(1))
        if a(1)[z - 1] = asc("/") then
          x = 1
          y -= 1
        end if
        if (a(1)[z - 1] >= asc("1")) andalso (a(1)[z - 1] <= asc("8")) then
          do while a(1)[z - 1] > asc("0")
            .brett(10 * x + y) = ohnewert
            a(1)[z - 1] -= 1
            x += 1
          loop
        end if
        if instr(symbol, chr(a(1)[z - 1])) > 0 then
          select case instr(symbol, chr(a(1)[z - 1])) - 7
            case -6 : .brett(10 * x + y) = - 1 * koenigswert
            case -5 : .brett(10 * x + y) = - 1 * damenwert
            case -4 : .brett(10 * x + y) = - 1 * turmwert
            case -3 : .brett(10 * x + y) = - 1 * laeuferwert
            case -2 : .brett(10 * x + y) = - 1 * springerwert
            case -1 : .brett(10 * x + y) = - 1 * bauernwert
            case +1 : .brett(10 * x + y) = + 1 * bauernwert
            case +2 : .brett(10 * x + y) = + 1 * springerwert
            case +3 : .brett(10 * x + y) = + 1 * laeuferwert
            case +4 : .brett(10 * x + y) = + 1 * turmwert
            case +5 : .brett(10 * x + y) = + 1 * damenwert
            case +6 : .brett(10 * x + y) = + 1 * koenigswert
          end select
          x += 1
        end if
      next z
      
      for y = 8 to 1 step -1
        for x = 1 to 8
          .material += .brett(10 * x + y)
        next x
      next y
      
      .farbe = iif(a(2) = "w", +1, -1)
      
      if len(a(4)) = 2 then
        .epfeld(-1 * .farbe) = 10 * (a(4)[0] - 96) + (a(4)[1] - 48)
      else
        .epfeld(-1 * .farbe) = 0
      end if
      .epfeld(.farbe) = 0
    end with
  
    halbzugzahl = 2 * (valint(a(6)) - 1) + iif(spielbrett.farbe = -1, 1, 0)
    
  end if
  
end sub ' Grundstellung

function tSchachprogramm.Brettdarstellung as string
' Zweck: Darstellung des Bretts auf dem Monitor
  dim as string s = _
  "  1 2 3 4 5 6 7 8 " & chr(13, 10) & _
  "8                 " & chr(13, 10) & _
  "7                 " & chr(13, 10) & _
  "6                 " & chr(13, 10) & _
  "5                 " & chr(13, 10) & _
  "4                 " & chr(13, 10) & _
  "3                 " & chr(13, 10) & _
  "2                 " & chr(13, 10) & _
  "1                 " & chr(13, 10) & _
  "  1 2 3 4 5 6 7 8 "
  for y as integer = 8 to 1 step -1
    for x as integer = 1 to 8
      select case abs(spielbrett.brett(10 * x + y))
        case ohnewert
          s[20 * (8 - y) + 2 * x + 20] = asc("-")
        case bauernwert
          s[20 * (8 - y) + 2 * x + 20] = asc("B")
        case springerwert
          s[20 * (8 - y) + 2 * x + 20] = asc("S")
        case laeuferwert
          s[20 * (8 - y) + 2 * x + 20] = asc("L")
        case turmwert
          s[20 * (8 - y) + 2 * x + 20] = asc("T")
        case damenwert
          s[20 * (8 - y) + 2 * x + 20] = asc("D")
        case koenigswert
          s[20 * (8 - y) + 2 * x + 20] = asc("K")
      end select
      if spielbrett.brett(10 * x + y) < 0 then
        s[20 * (8 - y) + 2 * x + 20] += 32
      end if
    next x
  next y
  function = s
end function

sub tSchachprogramm.Setzen(byref b as brt, byval von as integer, byval nach as integer)
' Zweck: Einen Zug auf dem Spielbrett ausfhren
  if (abs(nach - von) = 2) and (abs(b.brett(von)) = bauernwert) then
  ' Wenn Bauer zwei Felder zieht, dann En-Passant ermglichen
    b.epfeld(b.farbe) = von + bauer(1) * b.farbe
  else
    b.epfeld(b.farbe) = 0
  end if
  if (nach = b.epfeld(- 1 * b.farbe)) and (b.brett(von) = bauernwert * b.farbe) then
  ' Wenn En-Passant-Zug, dann erst den Bauern des Gegners ein Feld zurcksetzen
     b.epfeld(b.farbe) = 0
     Setzen(b, von, (nach \ 10) * 10 + von mod 10)
     b.farbe = - 1 * b.farbe
     von = (nach \ 10) * 10 + von mod 10
  end if
  if ((von >= 51) and (von <= 58)) _
  and ((nach = 71) _
    or (nach = 31) _
    or (nach = 78) _
    or (nach = 38)) _
  and (abs(b.brett(von)) = koenigswert) then
  ' Beginnen mit Rochade
    b.brett(nach) = b.brett(von)
  ' Knig setzen
    b.brett(von) = 0
    if nach \ 10 = 7 then
    ' kurze Rochade
      von = (von mod 10) + 80
      nach = von - 20
    ' Turm geht 2 Felder
    else
    ' lange Rochade
      von = (von mod 10) + 10
      nach = von + 30
    ' Turm geht 3 Felder
    end if
    b.brett(nach) = b.brett(von)
    b.brett(von) = 0
  else
  ' Ein ganz normaler Zug
    b.material = b.material - b.brett(nach)
    b.brett(nach) = b.brett(von)
    b.brett(von) = 0
    if ((nach mod 10 = 1) or (nach mod 10 = 8)) and (abs(b.brett(nach))) = bauernwert then
    ' Damenumwandlung
      b.brett(nach) = damenwert * b.farbe
      b.material = b.material + (damenwert - bauernwert) * b.farbe
    end if
  end if
  b.farbe = -1 * b.farbe
end sub ' Setzen

sub tSchachprogramm.Eintragen(byref b as brt, byval von as integer, byval nach as integer)
' Zweck: Einen gltigen Zug in die Zugliste eintragen
  if abs(b.material) > 83 then exit sub
' Weiterspielen hat keinen Zweck mehr, Ende
  b.anz += 1
  b.zuglistevon(b.anz) = von
  b.zuglistenach(b.anz) = nach
end sub ' Eintragen

sub tSchachprogramm.Zuggenerator_ohne(byref b as brt)
' Zweck: Alle mglichen Zgen ermitteln (ohne Rochade)
  dim as integer von, nach
  dim as integer i, j, k, l
  b.anz = 0 ' Anzahl der gefundenen Zge
  for j = 1 to 8 ' Alle Spielfelder durchlaufen
    for k = 1 to 8
      von = j * 10 + k ' von = 11 to 88
      if b.brett(von) <> ohnewert then
        select case b.brett(von) * b.farbe
        case bauernwert
          nach = von + bauer(1) * b.farbe
          if b.brett(nach) = ohnewert then
            eintragen(b, von, nach)
            if (b.farbe = 1) and (k = 2) or (b.farbe = -1) and (k = 7) then
              nach = von + bauer(2) * b.farbe
              if (b.brett(nach) = ohnewert) then
                Eintragen(b, von, nach)
              end if
            end if
          end if
          for i = 3 to 4
            nach = von + bauer(i) * b.farbe
            if ((-1 * b.brett(nach) * b.farbe >= bauernwert) _
            and (-1 * b.brett(nach) * b.farbe <= koenigswert)) _
            or (nach = b.epfeld(-1 * b.farbe)) then
              Eintragen(b, von, nach)
            end if
          next i
        case springerwert
          for i = 1 to 8
            nach = von + springer(i)
            if (-1 * b.brett(nach) * b.farbe >= ohnewert) _
            and (-1 * b.brett(nach) * b.farbe <= koenigswert) then
              Eintragen(b, von, nach)
            end if
          next i
        case laeuferwert
          for i = 1 to 4
            nach = von
            do
              nach += laeufer(i)
              if (-1 * b.brett(nach) * b.farbe >= ohnewert) _
              and (-1 * b.brett(nach) * b.farbe <= koenigswert) then
                Eintragen(b, von, nach)
              end if
            loop until b.brett(nach) <> ohnewert
          next i
        case turmwert
          for i = 1 to 4
            nach = von
            do
              nach += turm(i)
              if (-1 * b.brett(nach) * b.farbe >= ohnewert) _
              and (-1 * b.brett(nach) * b.farbe <= koenigswert) then
                Eintragen(b, von, nach)
              end if
            loop until b.brett(nach) <> ohnewert
          next i
        case koenigswert
          for i = 1 to 8
            nach = von + koenig(i)
            if (-1 * b.brett(nach) * b.farbe >= ohneWert) _
            and (-1 * b.brett(nach) * b.farbe <= Koenigswert) then
            ' Testen wo der gegnerische Knig steht
              l = 0
              do
                l += 1
              loop until (-1 * b.brett(nach + koenig(l)) * b.farbe = koenigswert) or (l = 8)
              if l = 8 then Eintragen(b, von, nach)
            end if
          next i
        case damenwert
          for i = 1 to 8
            nach = von
            do
              nach += koenig(i)
              if (-1 * b.brett(nach) * b.farbe >= ohnewert) _
              and (-1 * b.brett(nach) * b.farbe <= koenigswert) then
                Eintragen(b, von, nach)
              end if
            loop until b.brett(nach) <> ohnewert
          next i
        end select
      end if
    next k
  next j
end sub ' Zuggenerator_ohne

function tSchachprogramm.Schach(byref b as brt) as integer
' Zweck: Testen, ob der Knig im Schach steht
' Aufruf nachdem gesetzt wurde
  b.farbe = -1 * b.farbe
  zuggenerator_ohne(b)
  b.farbe = -1 * b.farbe
  for i as integer = 1 to b.anz
    if abs(b.brett(b.zuglistenach(i))) = koenigswert then
      return -1 ' wenn der Gegner den Knig erreicht
    end if
  next i
  return 0
end function ' Schach

sub tSchachprogramm.Rochade_testen(byref b as brt, byval vonkoenig as integer, byval schritt as integer, byval turmfeld as integer)
  dim brett1 as brt
  dim i as integer
  if b.brett(turmfeld) * b.farbe <> turmwert then exit sub
  i = vonkoenig + schritt
  do
    if b.brett(i) <> ohnewert then exit sub
    i += schritt
  loop until i = turmfeld
  brett1 = b
  if not Schach(brett1) then Eintragen(b, vonkoenig, vonkoenig + 2 * schritt)
' Steht der Knig auch nicht im Schach?
end sub

sub tSchachprogramm.Zuggenerator(byref b as brt)
' Zweck: Ermitteln aller Zge incl. Rochade
' erst normale Zge ermitteln
  Zuggenerator_ohne(b)
' danach die Rochadezge
  if b.brett(51) * b.farbe = koenigswert then
    Rochade_testen(b, 51, 10, 81)
    Rochade_testen(b, 51,-10, 11)
  end if
  if b.brett(58) * b.farbe = koenigswert then
    Rochade_testen(b, 58, 10, 88)
    Rochade_testen(b, 58,-10, 18)
  end if
end sub ' Zuggenerator

function tSchachprogramm.Maxwertung(byref b as brt, byval farbe as integer, byval tiefe as integer, byval alfa as integer) as integer
' Zweck: Rekursive Bewertung eines Spielbretts
  dim as integer minmax, wertung, zugnr
  dim as brt brett1
  dim as integer beta
  dim as integer alphabetaabbruch

  Zuggenerator(b) ' alle Zge incl. Rochade
  zugnr = 0
  beta = -32000 * b.farbe * farbe
  alphabetaabbruch = 0
  while (zugnr < b.anz) and not alphabetaabbruch
    zugnr += 1
    brett1 = b
    Setzen(brett1, b.zuglistevon(zugnr), b.zuglistenach(zugnr))

    if ((tiefe >= rechentiefe) and (b.brett(b.zuglistenach(zugnr)) = 0)) or (tiefe = maxtiefe) then
      wertung = Stellungswert(brett1, farbe)
      varianten += 1
    else
      wertung = Maxwertung(brett1, farbe, tiefe + 1, beta)
    end if

    if b.farbe = farbe then
      if wertung > beta then beta = wertung
      if beta > alfa then alphabetaabbruch = -1
    else
      if wertung < beta then beta = wertung
      if beta < alfa then alphabetaabbruch = -1
    end if

    b.zugwert(zugnr) = wertung
  wend
  Maxwertung = beta
end function ' Maxwertung

function tSchachprogramm.Spielereingabe(byref eingabe as string) as integer
' Zweck: Feststellen, was der Spieler will
  dim as integer ausgabe, zug, von, nach
  dim as brt brett1
  
  if (len(eingabe) = 4) andalso _
  (eingabe[0] >= asc("a")) and (eingabe[0] <= asc("h")) and _
  (eingabe[1] >= asc("1")) and (eingabe[1] <= asc("8")) and _
  (eingabe[2] >= asc("a")) and (eingabe[2] <= asc("h")) and _
  (eingabe[3] >= asc("1")) and (eingabe[3] <= asc("8")) then
    eingabe[0] -= (16 + 32)
    eingabe[2] -= (16 + 32)
  else
    return 0
  end if
  
  ausgabe = 0
  
  zug = valint(eingabe)
  Zuggenerator(spielbrett) ' Alle Zge incl. Rochade
  for i as integer = 1 to spielbrett.anz
    if zug = spielbrett.zuglistevon(i) * 100 + spielbrett.zuglistenach(i) then
      ausgabe = -1
      exit for
    end if
  next i
  
  if ausgabe = 0 then return 0
  
  von = zug \ 100
  nach = zug mod 100
  brett1 = spielbrett
  Setzen(brett1, von, nach)
  brett1.farbe = -1 * brett1.farbe
  if Schach(brett1) then
    ausgabe = 0
  else
    Setzen(spielbrett, von, nach)
    halbzugzahl += 1
  end if
  
  function = ausgabe
end function ' Spielereingabe

function tSchachprogramm.Zugausgeben(byval von as integer, byval nach as integer) as string
  dim zugstr as string
  dim zug as integer
  zug = von * 100 + nach
  zugstr = str(zug)
  zugstr[0] += 16 + 32
  zugstr[2] += 16 + 32
  function = zugstr
end function ' Zugausgeben

function tSchachprogramm.Feld_ist_in_grundstellung(byref b as brt, byval feldnr as integer) as integer
  if b.brett(feldnr) = grundbrett.brett(feldnr) then
    return -1
  else
    return 0
  end if
end function

function tSchachprogramm.Besterzug as integer
  dim brett1 as brt
  dim as integer i, j
  dim bestcounter as integer

  eom =0
  for i = 1 to spielbrett.anz
    if spielbrett.zugwert(i) = wert then
      eom += 1
      mgzugvon(eom) = spielbrett.zuglistevon(i)
      mgzugnach(eom) = spielbrett.zuglistenach(i)
      mgzugwert(eom) = 0
    end if
  next i

  bestcounter = 0
  Besterzug = 1

  for i =1 to eom

    brett1 = spielbrett
    mgzugwert(i) = 0

    if Feld_ist_in_grundstellung(brett1, mgzugvon(i)) then
      mgzugwert(i) += 5
      if abs(brett1.brett(mgzugvon(i))) = bauernwert then mgzugwert(i) += 2
    end if

    if (abs(brett1.brett(mgzugvon(i))) = koenigswert) then mgzugwert(i) -= 10

    if (halbzugzahl < 32) and ( _
      (abs(brett1.brett(mgzugvon(i))) = bauernwert) or _
      (abs(brett1.brett(mgzugvon(i))) = laeuferwert) or _
      (abs(brett1.brett(mgzugvon(i))) = springerwert) _
      ) then mgzugwert(i) += 20

    if (mgzugvon(i) \ 10 = 1) _
    or (mgzugvon(i) \ 10 = 8) _
    or (mgzugvon(i) mod 10 = 1) _
    or (mgzugvon(i) mod 10 = 8) then mgzugwert(i) += 2

    if (mgzugnach(i) \ 10 = 1) _
    or (mgzugnach(i) \ 10 = 8) _
    or (mgzugnach(i) mod 10 = 1) _
    or (mgzugnach(i) mod 10 = 8) then mgzugwert(i) -= 2

    brett1 = spielbrett
    Setzen(brett1, mgzugvon(i), mgzugnach(i))
    if Feld_ist_in_grundstellung(brett1, mgzugnach(i)) then mgzugwert(i) -= 10

    brett1.farbe = -1 * brett1.farbe
    Zuggenerator_ohne(brett1)

    for j = 1 to brett1.anz
      mgzugwert(i) += 1
      if brett1.zuglistenach(j) <> 0 then mgzugwert(i) += 1
    next j

    brett1.farbe = -1 * brett1.farbe
    Zuggenerator_ohne(brett1)

    for j = 1 to brett1.anz
      mgzugwert(i) -= 1
      if brett1.zuglistenach(j) <> 0 then mgzugwert(i) -= 1
    next j

    if mgzugwert(i) >= bestcounter then
      bestcounter = mgzugwert(i)
      Besterzug = i
    end if

  next i
end function ' Besterzug

function tSchachprogramm.Computerzug as string
' Zweck: Den besten Computerzug berechnen und ausfhren
  dim as integer i, schon_im_schach
  dim as string ausgabe

  varianten = 0

  wert = Maxwertung(spielbrett, spielbrett.farbe, 1, 32000)
  i = Besterzug

  halbzugzahl += 1

  schon_im_schach = Schach(spielbrett)
  
  Setzen(spielbrett, mgzugvon(i), mgzugnach(i))
  
  spielbrett.farbe = -1 * spielbrett.farbe

  if Schach(spielbrett) then
    ausgabe = "0000"
    if schon_im_schach then
      ausgabe &= "11" ' Schachmatt
    else
      ausgabe &= "10" ' Patt
    end if
    return ausgabe
  else
    ausgabe = Zugausgeben(mgzugvon(i), mgzugnach(i))
  end if
  
  spielbrett.farbe = -1 * spielbrett.farbe
  
  if Schach(spielbrett) then
    ausgabe &= "01" ' Schach
  else
    ausgabe &= "00"
  end if
  
  function = ausgabe
end function ' Computerzug
